.nolist
.include "m328pdef.inc" ; Define device ATmega328P
.list
;
; Exemplo com chamada de subrotina. Calcula o fatorial de NUM (1 byte, 6 no maximo) com chamadas recursivas
; a subrotina FACTORIAL. Salva o numero a fatorar na memoria em N e o resultado final em PROD (2 bytes, 720 no maximo)
; Exemplo de programao em assembly: uso de segmentos .CSEG e .DSEG, alocacao de variaveis
; na memoria, uso do stack para passagem de parametros

;.NOLIST					; Inclui o arquivo de definicoes (com listagem desabilitada).
;.INCLUDE "m328pdef.inc"	; Comentado porque O Studio6 inclui por default conforme o device escolhido
;.LIST						; (ver menu Project->Properties->Tool Chain->General->Include File

.DEF ACC_L = R0     ; renomeia registradores
.DEF ACC_H = R1
.DEF COUNT = R16
.DEF IAUX = R17

.EQU NUM = 3        ; Numero para o calculo do fatorial

.DSEG           ; Estruturas a seguir serao alocadas no DATA SEGMENT (i.e., na RAM)
N: .BYTE 1		; aloca 1 byte na memoria SRAM para a variavel N
PROD: .BYTE 2	; aloca 2 bytes para a variavel PROD


.CSEG       ; Estruturas a seguir serao alocadas no CODE SEGMENTO (i.e, flash de programa)

.ORG 0
   RJMP INIT    ; Primeira instrucao apos RESET


INIT:
   LDI IAUX, LOW(RAMEND)	; Inicializando o stack pointer (topo da ram)
   OUT SPL, IAUX
   LDI IAUX, HIGH(RAMEND)
   OUT SPH, IAUX            ; SPH:SPL <- RAMEND

   LDI COUNT, NUM			; COUNT <- NUM
   STS N, COUNT				; SRAM(N) <- COUNT
   PUSH COUNT				; Passagem de parametro (NUM) pela pilha
   CALL FACTORIAL			; Chamada da funcao. Resultado retorna em R1:R0
   POP COUNT				; Tira o parametro da pilha
   LDI XL, LOW(PROD)		; X <- endereco(PROD)
   LDI XH, HIGH(PROD)
   ST X+, ACC_L				; Grava o resultado em SRAM(PROD) (2 bytes)
   ST X, ACC_H

   SLEEP


.EQU NL = 2+1			; Offset do parametro N na pilha
.DEF NI = R18			; Redefine R18

.CSEG					; Codigo seguinte tambm sera alocado no CODE SEGMENT

; Funcao FACTORIAL(N)
; Calcula o fatorial de N (N<=6). N deve ser passado pela pilha.
; Retorna o resultado em R1:R0.
; Destroi: registradores 0, 1, 18, 30, 31

FACTORIAL:
    IN ZL, SPL      ; Z <- SP (2 bytes)
	IN ZH, SPH      ; Ponteiro Z (de 16 bits) e' composto por R31:R30
	LDD NI, Z+NL    ; NI <- parametro N
	CPI NI, 2       ; if (NI < 2)
	BRSH ELSE_BLK	; senao desvia para o bloco "else"	
	   CLR ACC_H	; Bloco "then"
       CLR ACC_L    ;   NI < 2: fatorial(NI) = 1
	   INC ACC_L	; R1:R0 <- 1
	RJMP AHEAD
ELSE_BLK:
       DEC NI			; Bloco "else": calcula N*FACTORIAL(N-1)
       PUSH NI			; passa N-1 pela pilha
	   CALL FACTORIAL   ; chamada recursiva: R1:R0 <- FACTORIAL(N-1)
	   POP NI			; retira o parametro da pilha (limpa a pilha)
	   INC NI			; re-incrementa NI para recuperar o valor de N
       MUL NI, ACC_L    ; R1:R0 <- N*R0
AHEAD:
	RET
